"""Added importHash

Peek Plugin Database Migration Script

Revision ID: 56d805072d2a
Revises: 3a9cddaf05e5
Create Date: 2017-10-15 12:05:26.795359

"""

# revision identifiers, used by Alembic.
revision = '56d805072d2a'
down_revision = '3a9cddaf05e5'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('InternalGroup', sa.Column('importHash', sa.String(length=50), nullable=True), schema='core_user')
    op.create_index('idx_InternalGroupTable_importHash', 'InternalGroup', ['importHash'], unique=False, schema='core_user')
    op.add_column('InternalUser', sa.Column('importHash', sa.String(length=50), nullable=True), schema='core_user')
    op.create_index('idx_InternalUserTable_importHash', 'InternalUser', ['importHash'], unique=False, schema='core_user')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('idx_InternalUserTable_importHash', table_name='InternalUser', schema='core_user')
    op.drop_column('InternalUser', 'importHash', schema='core_user')
    op.drop_index('idx_InternalGroupTable_importHash', table_name='InternalGroup', schema='core_user')
    op.drop_column('InternalGroup', 'importHash', schema='core_user')
    # ### end Alembic commands ###



from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_core_user._private.PluginNames import userPluginFilt, \
    userPluginObservableName
from vortex.handler.TupleDataObservableProxyHandler import TupleDataObservableProxyHandler


def makeTupleDataObservableProxy():
    return TupleDataObservableProxyHandler(observableName=userPluginObservableName,
                                           proxyToVortexName=peekServerName,
                                           additionalFilt=userPluginFilt)




import logging

from twisted.internet.defer import Deferred

from peek_core_user._private.server.controller.ImportController import \
    ImportController
from peek_core_user.server.UserImportApiABC import UserImportApiABC

logger = logging.getLogger(__name__)


class UserImportApi(UserImportApiABC):
    def __init__(self, importController: ImportController):
        self._importController = importController

    def shutdown(self):
        pass

    def importInternalUsers(self, importHash: str, usersEncodedPayload: bytes) -> Deferred:
        return self._importController.importInternalUsers(importHash, usersEncodedPayload)

    def importInternalGroups(self, importHash: str, groupsEncodedPayload: bytes) -> Deferred:
        return self._importController.importInternalGroups(importHash, groupsEncodedPayload)



import logging

from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler, \
    TuplesProviderABC

from peek_core_user._private.server.tuple_providers.UserLoginUiSettingTupleProvider import \
    UserLoginUiSettingTupleProvider
from peek_core_user._private.tuples.UserLoginUiSettingTuple import UserLoginUiSettingTuple
from peek_plugin_base.PeekVortexUtil import peekClientName
from peek_core_user._private.PluginNames import userPluginFilt, \
    userPluginObservableName
from peek_core_user._private.server.tuple_providers.GroupDetailTupleProvider import \
    GroupDetailTupleProvider
from peek_core_user._private.server.tuple_providers.UserLoggedInTupleProvider import \
    UserLoggedInTupleProvider
from peek_core_user._private.tuples.UserLoggedInTuple import UserLoggedInTuple
from peek_core_user.tuples.GroupDetailTuple import GroupDetailTuple
from peek_core_user.tuples.UserListItemTuple import \
    UserListItemTuple
from .tuple_providers.UserListItemTupleProvider import UserListItemTupleProvider

logger = logging.getLogger(__name__)


class Dummy(TuplesProviderABC):
    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(self, filt: dict, *args):
        return Payload(filt, tuples=[]).makePayloadEnvelope().toVortexMsg()


def makeTupleDataObservableHandler(dbSessionCreator, ourApi):
    observable = TupleDataObservableHandler(observableName=userPluginObservableName,
                                            additionalFilt=userPluginFilt,
                                            acceptOnlyFromVortex=peekClientName)

    observable.addTupleProvider(GroupDetailTuple.tupleName(),
                                GroupDetailTupleProvider(ourApi))

    observable.addTupleProvider(UserListItemTuple.tupleName(),
                                UserListItemTupleProvider(dbSessionCreator, ourApi))

    observable.addTupleProvider(UserLoggedInTuple.tupleType(),
                                UserLoggedInTupleProvider(dbSessionCreator, ourApi))

    observable.addTupleProvider(UserLoginUiSettingTuple.tupleType(),
                                UserLoginUiSettingTupleProvider(dbSessionCreator))

    return observable



""" 
 *  Copyright Synerty Pty Ltd 2016
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""
import logging

from sqlalchemy import Column, DateTime
from sqlalchemy import Integer, String
from vortex.Tuple import Tuple, addTupleType

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class UserLoggedIn(Tuple, DeclarativeBase):
    """ UserLoggedIn

    This table stores which users are logged into which devices

    """
    __tupleType__ = userPluginTuplePrefix + 'UserLoggedIn'
    __tablename__ = 'UserLoggedIn'

    id = Column(Integer, primary_key=True, autoincrement=True)
    loggedInDateTime = Column(DateTime(True), nullable=False)
    userName = Column(String(50), unique=True)
    deviceToken = Column(String(100), unique=True, nullable=False)
    vehicle = Column(String)



import logging
from typing import List, Optional

from sqlalchemy import or_
from sqlalchemy.orm import subqueryload
from sqlalchemy.orm.exc import NoResultFound

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_core_user._private.storage.InternalGroupTuple import InternalGroupTuple
from peek_core_user._private.storage.InternalUserGroupTuple import \
    InternalUserGroupTuple
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.storage.UserLoggedIn import UserLoggedIn
from peek_core_user.server.UserInfoApiABC import UserInfoApiABC
from peek_core_user.tuples.GroupDetailTuple import GroupDetailTuple
from peek_core_user.tuples.UserDetailTuple import UserDetailTuple
from vortex.DeferUtil import deferToThreadWrapWithLogger

logger = logging.getLogger(__name__)


class UserInfoApi(UserInfoApiABC):
    #: A reference to the core device plugins API
    _deviceApi: DeviceApiABC

    _userCopyFields = (
            set(UserDetailTuple.tupleFieldNames())
            & set(InternalUserTuple.tupleFieldNames())
    )

    _groupCopyFields = (
            set(GroupDetailTuple.tupleFieldNames())
            & set(InternalGroupTuple.tupleFieldNames())
    )

    def __init__(self, deviceApi: DeviceApiABC, dbSessionCreator):
        self._deviceApi = deviceApi
        self._dbSessionCreator = dbSessionCreator

    def shutdown(self):
        pass

    @deferToThreadWrapWithLogger(logger)
    def user(self, userName: str):
        """
        :returns a Deferred firing with Optional[UserDetailTuple]]
        """
        return self.userBlocking(userName)

    def userBlocking(self, userName, ormSession=None) -> Optional[UserDetailTuple]:

        if ormSession:
            close = False

        else:
            ormSession = self._dbSessionCreator()
            close = True

        try:
            user = (ormSession.query(InternalUserTuple)
                    .filter(InternalUserTuple.userName == userName)
                    .one())

            return self._makeUserDetails(user)

        except NoResultFound:
            return None

        finally:
            if close:
                ormSession.close()

    @deferToThreadWrapWithLogger(logger)
    def users(self, likeTitle: Optional[str] = None,
              groupNames: Optional[List[str]] = None):
        """
        Returns Deferred[List[UserDetailTuple]]
        """
        ormSession = self._dbSessionCreator()
        try:
            return self.usersBlocking(ormSession, likeTitle, groupNames)
        finally:
            ormSession.close()

    def usersBlocking(self, session, likeTitle: Optional[str] = None,
                      groupNames: Optional[List[str]] = None) -> List[
        UserDetailTuple]:
        qry = (
            session
                .query(InternalUserTuple)
                .options(subqueryload(InternalUserTuple.groups))
                .order_by(InternalUserTuple.userName)
        )

        if groupNames:
            qry = (
                qry.join(InternalUserGroupTuple)
                    .join(InternalGroupTuple)
                    .filter(InternalGroupTuple.groupName.in_(groupNames))
            )

        if likeTitle:
            qry = qry.filter(or_(
                InternalUserTuple.userTitle.ilike('%' + likeTitle + '%'),
                InternalUserTuple.userName.ilike('%' + likeTitle + '%')
            ))

        ormUsers = qry.all()

        return [self._makeUserDetails(u) for u in ormUsers]

    def _makeUserDetails(self, ormUser: InternalUserTuple) -> UserDetailTuple:
        userDetail = UserDetailTuple()
        for fieldName in self._userCopyFields:
            setattr(userDetail, fieldName, getattr(ormUser, fieldName))

        userDetail.groupNames = [g.groupName for g in ormUser.groups]
        return userDetail

    @deferToThreadWrapWithLogger(logger)
    def groups(self, likeTitle: Optional[str] = None):
        """
        Returns Deferred[List[UserDetailTuple]]
        """
        ormSession = self._dbSessionCreator()
        try:
            return self.groupsBlocking(ormSession, likeTitle)
        finally:
            ormSession.close()

    def groupsBlocking(self, session, likeTitle: Optional[str] = None) -> List[
        GroupDetailTuple]:
        qry = (
            session
                .query(InternalGroupTuple)
                .order_by(InternalGroupTuple.groupName)
        )

        if likeTitle:
            qry = qry.filter(InternalGroupTuple.userTitle.ilike('%' + likeTitle + '%'))

        return [self._makeGroupDetails(u) for u in qry.all()]

    def _makeGroupDetails(self, ormGroups: InternalGroupTuple) -> GroupDetailTuple:
        groupDetail = GroupDetailTuple()
        for fieldName in self._groupCopyFields:
            setattr(groupDetail, fieldName, getattr(ormGroups, fieldName))

        return groupDetail

    @deferToThreadWrapWithLogger(logger)
    def peekDeviceTokenForUser(self, userName: str) -> Optional[str]:
        session = self._dbSessionCreator()
        try:
            result = (session.query(UserLoggedIn)
                      .filter(UserLoggedIn.userName == userName)
                      .one())

            return result.deviceToken

        except NoResultFound:
            return None

        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def peekUserForDeviceToken(self, deviceToken) -> Optional[UserDetailTuple]:
        session = self._dbSessionCreator()
        try:
            result = (
                session
                    .query(InternalUserTuple)
                    .join(UserLoggedIn,
                          UserLoggedIn.userName == InternalUserTuple.userName)
                    .filter(UserLoggedIn.deviceToken == deviceToken)
                    .one()
            )

            return self._makeUserDetails(result)

        except NoResultFound:
            return None

        finally:
            session.close()



import logging

from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_plugin_base.PeekVortexUtil import peekAdminName
from peek_core_user._private.PluginNames import userPluginFilt, \
    userPluginObservableName
from peek_core_user._private.server.admin_tuple_providers.LoggedInUserStatusTupleProvider import \
    LoggedInUserStatusTupleProvider
from peek_core_user._private.server.tuple_providers.GroupDetailTupleProvider import \
    GroupDetailTupleProvider
from peek_core_user._private.server.tuple_providers.UserListItemTupleProvider import \
    UserListItemTupleProvider
from peek_core_user._private.server.tuple_providers.UserLoggedInTupleProvider import \
    UserLoggedInTupleProvider
from peek_core_user._private.tuples.LoggedInUserStatusTuple import \
    LoggedInUserStatusTuple
from peek_core_user._private.tuples.UserLoggedInTuple import UserLoggedInTuple
from peek_core_user.tuples.GroupDetailTuple import GroupDetailTuple
from peek_core_user.tuples.UserListItemTuple import UserListItemTuple

logger = logging.getLogger(__name__)


def makeAdminTupleDataObservableHandler(dbSessionCreator, deviceApi: DeviceApiABC, ourApi):
    observable = TupleDataObservableHandler(observableName=userPluginObservableName,
                                            additionalFilt=userPluginFilt,
                                            acceptOnlyFromVortex=peekAdminName)

    observable.addTupleProvider(
        LoggedInUserStatusTuple.tupleName(),
        LoggedInUserStatusTupleProvider(dbSessionCreator, deviceApi)
    )

    observable.addTupleProvider(GroupDetailTuple.tupleName(),
                                GroupDetailTupleProvider(ourApi))

    observable.addTupleProvider(UserListItemTuple.tupleName(),
                                UserListItemTupleProvider(dbSessionCreator, ourApi))

    observable.addTupleProvider(UserLoggedInTuple.tupleType(),
                                UserLoggedInTupleProvider(dbSessionCreator, ourApi))

    return observable



"""Added vehical to LoggedIn

Peek Plugin Database Migration Script

Revision ID: 45c62f8acd87
Revises: 2c6728ebc562
Create Date: 2018-12-08 19:21:47.019204

"""

# revision identifiers, used by Alembic.
revision = '45c62f8acd87'
down_revision = '2c6728ebc562'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('UserLoggedIn', sa.Column('vehicle', sa.String(), nullable=True), schema='core_user')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('UserLoggedIn', 'vehicle', schema='core_user')
    # ### end Alembic commands ###



from typing import List

from vortex.Tuple import addTupleType, TupleField
from vortex.TupleAction import TupleActionABC

from peek_core_user._private.PluginNames import userPluginTuplePrefix


@addTupleType
class UserLoginAction(TupleActionABC):
    __tupleType__ = userPluginTuplePrefix + "UserLoginAction"
    userName: str = TupleField()
    password: str = TupleField()
    deviceToken: str = TupleField()
    vehicleId: str = TupleField()

    #: A list of accepted warning keys
    # If any server side warnings occur and they are in this list then the logon
    # continues.
    acceptedWarningKeys: List[str] = TupleField()



import logging

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from vortex.Tuple import addTupleType, Tuple, TupleField

logger = logging.getLogger(__name__)


@addTupleType
class UserListItemTuple(Tuple):
    __tupleType__ = userPluginTuplePrefix + "UserListItemTuple"

    #:  The unique ID of the user
    userId: str = TupleField()

    #:  The nice name of the user
    displayName: str = TupleField()

    @property
    def userName(self) -> str:
        return self.userId

    @property
    def userTitle(self) -> str:
        return self.displayName



from typing import Callable

from abc import ABCMeta, abstractmethod
from twisted.internet.defer import Deferred

from peek_core_user.server.UserHookApiABC import UserHookApiABC
from peek_core_user.server.UserImportApiABC import UserImportApiABC
from peek_core_user.server.UserInfoApiABC import UserInfoApiABC
from peek_core_user.server.UserLoginApiABC import UserLoginApiABC
from peek_core_user.tuples.login.UserLoginResponseTuple import UserLoginResponseTuple
from peek_core_user.tuples.login.UserLogoutResponseTuple import UserLogoutResponseTuple

UserPostLoginHookCallable = Callable[[UserLoginResponseTuple], Deferred]

UserPostLogoutHookCallable = Callable[[UserLogoutResponseTuple], Deferred]


class UserApiABC(metaclass=ABCMeta):

    @property
    @abstractmethod
    def loginApi(self) -> UserLoginApiABC:
        """ Login API

        Returns the API class that handles the user logins

        :return A reference to the UserLoginApiABC class
        """

    @property
    @abstractmethod
    def importApi(self) -> UserImportApiABC:
        """ Import API

        Returns the API class that can be used to import the internal user/group objects

        :return A reference to the UserImportApiABC class
        """

    @property
    @abstractmethod
    def hookApi(self) -> UserHookApiABC:
        """ Hook API

        Returns the API class that can be used to attach callbacks and observers to the
        login/logout process

        :return A reference to the UserHookApiABC class
        """

    @property
    @abstractmethod
    def infoApi(self) -> UserInfoApiABC:
        """ Info API

        Returns the API class that provides information about users

        :return A reference to the UserInfoApiABC class
        """



"""Implemented internal directory

Peek Plugin Database Migration Script

Revision ID: 0498d92479c8
Revises: 3fec57d8da58
Create Date: 2017-10-07 17:33:08.316831

"""

# revision identifiers, used by Alembic.
revision = '0498d92479c8'
down_revision = '3fec57d8da58'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.execute('TRUNCATE TABLE core_user."UserLoggedIn" ')

    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('InternalGroup',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('groupName', sa.String(length=50), nullable=False),
    sa.Column('groupTitle', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('groupName'),
    sa.UniqueConstraint('groupTitle'),
    schema='core_user'
    )
    op.create_table('InternalUser',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('userName', sa.String(length=50), nullable=False),
    sa.Column('userTitle', sa.String(length=50), nullable=False),
    sa.Column('userUuid', sa.String(length=50), nullable=False),
    sa.Column('password', sa.String(length=50), nullable=False),
    sa.Column('mobile', sa.String(length=50), nullable=True),
    sa.Column('email', sa.String(length=50), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('userName'),
    sa.UniqueConstraint('userTitle'),
    sa.UniqueConstraint('userUuid'),
    schema='core_user'
    )
    op.create_table('InternalUserGroup',
    sa.Column('userId', sa.Integer(), nullable=False),
    sa.Column('groupId', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['groupId'], ['core_user.InternalGroup.id'], ondelete='CASCADE'),
    sa.ForeignKeyConstraint(['userId'], ['core_user.InternalUser.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('userId', 'groupId'),
    schema='core_user'
    )
    op.create_index('idx_InternalUserGroup_map', 'InternalUserGroup', ['userId', 'groupId'], unique=True, schema='core_user')
    op.add_column('UserLoggedIn', sa.Column('deviceToken', sa.String(length=100), nullable=False), schema='core_user')
    op.add_column('UserLoggedIn', sa.Column('userName', sa.String(length=50), nullable=True), schema='core_user')
    op.drop_constraint('UserLoggedIn_clientDeviceToken_key', 'UserLoggedIn', schema='core_user', type_='unique')
    op.drop_constraint('UserLoggedIn_userId_key', 'UserLoggedIn', schema='core_user', type_='unique')
    op.create_unique_constraint(None, 'UserLoggedIn', ['userName'], schema='core_user')
    op.create_unique_constraint(None, 'UserLoggedIn', ['deviceToken'], schema='core_user')
    op.drop_column('UserLoggedIn', 'userId', schema='core_user')
    op.drop_column('UserLoggedIn', 'clientDeviceToken', schema='core_user')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('UserLoggedIn', sa.Column('clientDeviceToken', sa.VARCHAR(length=100), autoincrement=False, nullable=False), schema='core_user')
    op.add_column('UserLoggedIn', sa.Column('userId', sa.VARCHAR(length=50), autoincrement=False, nullable=True), schema='core_user')
    op.drop_constraint(None, 'UserLoggedIn', schema='core_user', type_='unique')
    op.drop_constraint(None, 'UserLoggedIn', schema='core_user', type_='unique')
    op.create_unique_constraint('UserLoggedIn_userId_key', 'UserLoggedIn', ['userId'], schema='core_user')
    op.create_unique_constraint('UserLoggedIn_clientDeviceToken_key', 'UserLoggedIn', ['clientDeviceToken'], schema='core_user')
    op.drop_column('UserLoggedIn', 'userName', schema='core_user')
    op.drop_column('UserLoggedIn', 'deviceToken', schema='core_user')
    op.drop_index('idx_InternalUserGroup_map', table_name='InternalUserGroup', schema='core_user')
    op.drop_table('InternalUserGroup', schema='core_user')
    op.drop_table('InternalUser', schema='core_user')
    op.drop_table('InternalGroup', schema='core_user')
    # ### end Alembic commands ###



""" 
 *  Copyright Synerty Pty Ltd 2017
 *
 *  MIT License
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""

import logging

from sqlalchemy import Column
from sqlalchemy import Integer, String, Index
from sqlalchemy.orm import relationship
from vortex.Tuple import Tuple, addTupleType, TupleField

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase
from peek_core_user._private.storage.InternalGroupTuple import InternalGroupTuple
from peek_core_user._private.storage.InternalUserGroupTuple import \
    InternalUserGroupTuple

logger = logging.getLogger(__name__)


@addTupleType
class InternalUserTuple(Tuple, DeclarativeBase):
    """ Internal

    This table doesn't do anything

    """
    __tupleType__ = userPluginTuplePrefix + 'InternalUserTuple'
    __tablename__ = 'InternalUser'

    id = Column(Integer, primary_key=True, autoincrement=True)
    userName = Column(String, unique=True, nullable=False)
    userTitle = Column(String, unique=True, nullable=False)
    userUuid = Column(String, unique=True, nullable=False)

    importHash = Column(String)

    mobile = Column(String)
    email = Column(String)

    groups = relationship(InternalGroupTuple, secondary=InternalUserGroupTuple.__table__)

    #: This field is ussed for the admin-app to edit the groups
    groupIds = TupleField()

    __table_args__ = (
        Index("idx_InternalUserTable_importHash", importHash),
    )



import logging
from datetime import datetime
from typing import List

import pytz
from txcelery.defer import DeferrableTask

from peek_plugin_base.worker import CeleryDbConn
from peek_core_user._private.storage.InternalGroupTuple import InternalGroupTuple
from peek_core_user._private.tuples.InternalGroupImportResultTuple import \
    InternalGroupImportResultTuple
from peek_core_user._private.worker.CeleryApp import celeryApp
from peek_core_user.tuples.import_.ImportInternalGroupTuple import \
    ImportInternalGroupTuple
from vortex.Payload import Payload

logger = logging.getLogger(__name__)


@DeferrableTask
@celeryApp.task(bind=True)
def importInternalGroups(self, importHash:str, groupsVortexMsg: bytes) -> InternalGroupImportResultTuple:
    """ Import Internal Groups

    :param self: A celery reference to this task
    :param importHash: An unique string of this group of items being imported.
    :param groupsVortexMsg: A vortexMsg containing the group tuples
    :returns: A list of grid keys that have been updated.
    """

    importGroups: List[ImportInternalGroupTuple] = (
        Payload()
            .fromEncodedPayload(groupsVortexMsg)
            .tuples
    )

    startTime = datetime.now(pytz.utc)

    session = CeleryDbConn.getDbSession()
    try:
        sameCount = 0
        updates = []
        deleteIds = []
        inserts = []
        errors = []

        # This will remove duplicates
        allNames = [i.groupName for i in importGroups]

        existingGroupsByName = {
            g.groupName: g for g in
            session
                .query(InternalGroupTuple)
                .filter(InternalGroupTuple.userName.in_(allNames))
                .filter(InternalGroupTuple.importHash == importHash)
                .all()
        }

        for importGroup in importGroups:
            existingGroup = existingGroupsByName.pop(importGroup.grouName, None)
            if existingGroup:
                updated = False
                for fieldName in ImportInternalGroupTuple.tupleFieldNames():
                    newVal = getattr(importGroup, fieldName)
                    if getattr(existingGroup, fieldName) != newVal:
                        setattr(existingGroup, fieldName, newVal)
                        updated = True

                if updated:
                    updates.append(existingGroup)
                else:
                    sameCount += 1

            else:
                newGroup = InternalGroupTuple()

                for fieldName in ImportInternalGroupTuple.tupleFieldNames():
                    setattr(newGroup, fieldName, getattr(importGroup, fieldName))

                session.add(newGroup)
                inserts.append(newGroup)

        for oldGroup in existingGroupsByName.values():
            deleteIds.append(oldGroup.id)
            session.delete(oldGroup)

        session.commit()
        logger.info("Inserted %s, Updated %s, Deleted %s, Same %s, in %s",
                    len(inserts), len(updates), len(deleteIds), sameCount,
                    (datetime.now(pytz.utc) - startTime))

        return InternalGroupImportResultTuple(
            addedIds=[o.id for o in inserts],
            updatedIds=[o.id for o in updates],
            deletedIds=deleteIds,
            sameCount=sameCount,
            errors=errors
        )

    except Exception as e:
        session.rollback()
        logger.warning("Task failed, but it will retry. %s", e)
        raise self.retry(exc=e, countdown=2)

    finally:
        session.close()



from typing import Optional, List

from abc import ABCMeta, abstractmethod
from twisted.internet.defer import Deferred

from peek_core_user.tuples.UserDetailTuple import UserDetailTuple


class UserInfoApiABC(metaclass=ABCMeta):
    @abstractmethod
    def user(self, userName: str) -> Deferred:
        """ Users

        :param userName: The userName of the user to retrieve

        :return: A Deferred, firing with Optional[UserDetailTuple]
        """

    @abstractmethod
    def userBlocking(self, userName:str, ormSession=None) -> Optional[UserDetailTuple]:
        """ User Details for User ID

        Return an instance of c{UserDetailTuple} for the userName provided.
        :param userName: The username to retrive the details for
        :param ormSession: Specify the ormSession to use, otherwise it may close the
                            current session.

        :return UserDetailTuple
        """

    @abstractmethod
    def users(self, likeTitle: Optional[str] = None,
              groupNames: Optional[List[str]] = None) -> Deferred:
        """ Users

        :param likeTitle: An optional string to look for in the title of the users
        :param groupNames: An optional list of group names to include users for.

        :return: A Deferred, callbacking with a List[UserDetailTuple]
        """

    @abstractmethod
    def groups(self, likeTitle: Optional[str] = None) -> Deferred:
        """ Groups

        :param likeTitle: An optional string to look for in the title of the groups

        :return: A Deferred, callbacking with a List[GroupDetailTuple]
        """

    @abstractmethod
    def peekDeviceTokenForUser(self, userName:str) -> Deferred:  # [Optional[str]]:
        """Peek Device Token for Logged in User

        Return the first peek device token for a device this user is logged in to.
        Otherwise, return None

        :return Peek Device Token or None
        """


    @abstractmethod
    def peekUserForDeviceToken(self, deviceToken) -> Deferred:  #Optional[UserDetailTuple]:
        """Peek User for Device Token

        Return a user detail tuple for a user logged into a device with deviceToken

        :return UserDetail or None
        """


from peek_core_user._private.server.admin_backend.InternalGroupTableHandler import \
    makeInternalGroupTableHandler
from .InternalUserTableHandler import makeInternalUserTableHandler
from .SettingPropertyHandler import makeSettingPropertyHandler
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler


def makeAdminBackendHandlers(tupleObservable: TupleDataObservableHandler,
                             dbSessionCreator):
    yield makeInternalUserTableHandler(tupleObservable, dbSessionCreator)
    yield makeInternalGroupTableHandler(tupleObservable, dbSessionCreator)

    yield makeSettingPropertyHandler(tupleObservable, dbSessionCreator)




"""Added settings table

Peek Plugin Database Migration Script

Revision ID: 3a9cddaf05e5
Revises: 0498d92479c8
Create Date: 2017-10-14 14:58:46.523021

"""

# revision identifiers, used by Alembic.
revision = '3a9cddaf05e5'
down_revision = '0498d92479c8'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('Setting',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    schema='core_user'
    )
    op.create_table('SettingProperty',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('settingId', sa.Integer(), nullable=False),
    sa.Column('key', sa.String(length=50), nullable=False),
    sa.Column('type', sa.String(length=16), nullable=True),
    sa.Column('int_value', sa.Integer(), nullable=True),
    sa.Column('char_value', sa.String(), nullable=True),
    sa.Column('boolean_value', sa.Boolean(), nullable=True),
    sa.ForeignKeyConstraint(['settingId'], ['core_user.Setting.id'], ),
    sa.PrimaryKeyConstraint('id'),
    schema='core_user'
    )
    op.drop_index('idx_InternalUserGroup_map', table_name='InternalUserGroup', schema='core_user')
    op.create_index('idx_SettingProperty_settingId', 'SettingProperty', ['settingId'], unique=False, schema='core_user')
    op.create_index('idx_InternalUserGroup_map', 'InternalUserGroup', ['userId', 'groupId'], unique=True, schema='core_user')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('idx_InternalUserGroup_map', table_name='InternalUserGroup', schema='core_user')
    op.drop_index('idx_SettingProperty_settingId', table_name='SettingProperty', schema='core_user')
    op.drop_table('SettingProperty', schema='core_user')
    op.drop_table('Setting', schema='core_user')
    op.create_index('idx_InternalUserGroup_map', 'InternalUserGroup', ['userId', 'groupId'], unique=True, schema='core_user')
    # ### end Alembic commands ###



"""Removed string limits

Peek Plugin Database Migration Script

Revision ID: 3fd59734e444
Revises: 45c62f8acd87
Create Date: 2019-06-07 09:42:17.692382

"""

# revision identifiers, used by Alembic.
revision = '3fd59734e444'
down_revision = '45c62f8acd87'
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('idx_InternalUserPassword_userId', 'InternalUserPassword', ['userId'],
                    unique=True, schema='core_user')
    # ### end Alembic commands ###

    op.alter_column('InternalGroup', 'groupName', type_=sa.String(), nullable=False,
                    schema='core_user')
    op.alter_column('InternalGroup', 'importHash', type_=sa.String(), nullable=True,
                    schema='core_user')

    op.alter_column('InternalUserPassword', 'password', type_=sa.String(), nullable=False,
                    schema='core_user')

    op.alter_column('InternalUser', 'userName', type_=sa.String(), nullable=False,
                    schema='core_user')
    op.alter_column('InternalUser', 'userTitle', type_=sa.String(), nullable=False,
                    schema='core_user')
    op.alter_column('InternalUser', 'userUuid', type_=sa.String(), nullable=False,
                    schema='core_user')
    op.alter_column('InternalUser', 'importHash', type_=sa.String(), nullable=True,
                    schema='core_user')
    op.alter_column('InternalUser', 'mobile', type_=sa.String(), nullable=True,
                    schema='core_user')
    op.alter_column('InternalUser', 'email', type_=sa.String(), nullable=True,
                    schema='core_user')


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('idx_InternalUserPassword_userId', table_name='InternalUserPassword',
                  schema='core_user')
    # ### end Alembic commands ###



import logging

from peek_plugin_base.client.PluginClientEntryHookABC import PluginClientEntryHookABC
from peek_core_user._private.client.TupleActionProcessorProxy import \
    makeTupleActionProcessorProxy
from peek_core_user._private.client.TupleDataObservableProxy import \
    makeTupleDataObservableProxy

logger = logging.getLogger(__name__)


class PluginClientEntryHook(PluginClientEntryHookABC):
    def __init__(self, *args, **kwargs):
        PluginClientEntryHookABC.__init__(self, *args, **kwargs)

        self._handlers = []

    def load(self):
        # Load the tuples, so that Payload can parse them
        from peek_core_user._private.tuples import loadPrivateTuples
        loadPrivateTuples()

        from peek_core_user._private.storage.DeclarativeBase import loadStorageTuples
        loadStorageTuples()

        from peek_core_user.tuples import loadPublicTuples
        loadPublicTuples()

        logger.debug("loaded")

    def start(self):
        self._handlers.append(makeTupleDataObservableProxy())
        self._handlers.append(makeTupleActionProcessorProxy())

        logger.debug("starting")

    def stop(self):
        for handler in self._handlers:
            handler.shutdown()

        logger.debug("stopped")

    def unload(self):
        logger.debug("unloaded")




from peek_core_user._private.PluginNames import userPluginTuplePrefix
from vortex.Tuple import addTupleType, Tuple, TupleField


@addTupleType
class GroupDetailTuple(Tuple):
    __tupleType__ = userPluginTuplePrefix + "GroupDetailTuple"

    #:  ID
    id: int = TupleField()

    #:  The name of the group, EG C917
    groupName: str = TupleField()

    #:  The title of the group, EG 'Chief Wiggum'
    groupTitle: str = TupleField()



import logging

from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_core_user._private.storage.Setting import globalSetting, SHOW_LOGIN_AS_LIST, \
    SHOW_VEHICLE_INPUT
from peek_core_user._private.tuples.UserLoginUiSettingTuple import UserLoginUiSettingTuple
from peek_plugin_base.storage.DbConnection import DbSessionCreator

logger = logging.getLogger(__name__)


class UserLoginUiSettingTupleProvider(TuplesProviderABC):
    def __init__(self, dbSessionCreator: DbSessionCreator):
        self._dbSessionCreator = dbSessionCreator

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(self, filt: dict, tupleSelector: TupleSelector) -> Deferred:
        tuple_ = UserLoginUiSettingTuple()

        dbSession = self._dbSessionCreator()
        try:
            tuple_.showUsersAsList = globalSetting(dbSession, SHOW_LOGIN_AS_LIST)
            tuple_.showVehicleInput = globalSetting(dbSession, SHOW_VEHICLE_INPUT)

        finally:
            dbSession.close()

        payloadEnvelope = Payload(filt=filt, tuples=[tuple_]).makePayloadEnvelope()
        vortexMsg = payloadEnvelope.toVortexMsg()
        return vortexMsg



import logging
import uuid
from datetime import datetime
from typing import List

import pytz
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import subqueryload
from txcelery.defer import DeferrableTask

from peek_plugin_base.worker import CeleryDbConn
from peek_core_user._private.server.controller.PasswordUpdateController import \
    PasswordUpdateController
from peek_core_user._private.storage.InternalGroupTuple import InternalGroupTuple
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.tuples.InternalUserImportResultTuple import \
    InternalUserImportResultTuple
from peek_core_user._private.worker.CeleryApp import celeryApp
from peek_core_user.tuples.import_.ImportInternalUserTuple import \
    ImportInternalUserTuple
from vortex.Payload import Payload

logger = logging.getLogger(__name__)


@DeferrableTask
@celeryApp.task(bind=True)
def importInternalUsers(self, importHash: str,
                        usersVortexMsg: bytes) -> InternalUserImportResultTuple:
    """ Import Internal Users

    :param self: A celery reference to this task
    :param importHash: An unique string of this group of items being imported.
    :param usersVortexMsg: A vortexMsg containing the user tuples
    :returns: A list of grid keys that have been updated.
    """

    importUsers: List[ImportInternalUserTuple] = (
        Payload()
            .fromEncodedPayload(usersVortexMsg)
            .tuples
    )

    startTime = datetime.now(pytz.utc)

    session = CeleryDbConn.getDbSession()
    try:
        same = []
        updates = []
        deleteIds = []
        inserts = []
        errors = []

        # This will remove duplicates
        allNames = [i.userName for i in importUsers]

        if not allNames:
            existingUsersByName = {}

        else:
            existingUsersByName = {
                g.userName: g for g in
                session
                    .query(InternalUserTuple)
                    .filter(InternalUserTuple.userName.in_(allNames))
                    .filter(InternalUserTuple.importHash == importHash)
                    .options(subqueryload(InternalUserTuple.groups))
                    .all()
            }

        groupsByName = {
            g.groupName: g for g in
            session.query(InternalGroupTuple).all()
        }

        for importUser in importUsers:
            try:
                existingUser = existingUsersByName.pop(importUser.userName, None)
                if existingUser:
                    _updateUser(existingUser, groupsByName, importUser, same, updates)

                else:
                    _insertUser(session, groupsByName, importUser, importHash, inserts)

                session.commit()

            except IntegrityError as e:
                errors.append(str(e))
                session.rollback()

        for oldUser in existingUsersByName.values():
            deleteIds.append(oldUser.id)
            session.delete(oldUser)

        session.commit()

        logger.debug("Inserted %s, Updated %s, Deleted %s, Same %s, in %s",
                    len(inserts), len(updates), len(deleteIds), len(same),
                    (datetime.now(pytz.utc) - startTime))

        return InternalUserImportResultTuple(
            addedIds=[o.id for o in inserts],
            updatedIds=[o.id for o in updates],
            deletedIds=deleteIds,
            sameCount=len(same),
            errors=errors
        )

    except Exception as e:
        session.rollback()
        logger.warning("Task failed, but it will retry. %s", e)
        raise self.retry(exc=e, countdown=2)

    finally:
        session.close()


def _insertUser(session, groupsByName, importUser, importHash, inserts):

    newUser = InternalUserTuple()
    newUser.importHash = importHash

    for fieldName in ImportInternalUserTuple.tupleFieldNames():
        setattr(newUser, fieldName, getattr(importUser, fieldName))

    if importUser.groupKeys is not None:
        for groupKey in importUser.groupKeys:
            newUser.groups.append(groupsByName[groupKey])

    newUser.password = PasswordUpdateController.hashPass(str(uuid.uuid4()))

    session.add(newUser)
    inserts.append(newUser)


def _updateUser(existingUser, groupsByName, importUser, same, updates):

    excludeFieldNames = ("groupKeys", "password")

    copyFields = filter(lambda f: f not in excludeFieldNames,
                        ImportInternalUserTuple.tupleFieldNames())

    updated = False
    for fieldName in copyFields:
        newVal = getattr(importUser, fieldName)
        if getattr(existingUser, fieldName) != newVal:
            setattr(existingUser, fieldName, newVal)
            updated = True

    # The password is an optional field
    if importUser.password is not None:
        existingUser.password = PasswordUpdateController.hashPass(importUser.password)
        updated = True

    # If there are NONE groups, then don't make any changes
    if importUser.groupKeys is not None:
        linkedGroupNames = set([g.groupName for g in existingUser.groups])
        addGroups = set(importUser.groupKeys) - linkedGroupNames
        removeGroups = linkedGroupNames - set(importUser.groupKeys)

        for addGroup in addGroups:
            existingUser.groups.append(groupsByName[addGroup])

        for removeGroup in removeGroups:
            existingUser.groups.remove(groupsByName[removeGroup])

        updated = updated or bool(addGroups) or bool(removeGroups)


    if updated:
        updates.append(existingUser)
    else:
        same.append(existingUser)



""" 
 *  Copyright Synerty Pty Ltd 2017
 *
 *  MIT License
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""

import logging

from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy.sql.schema import Index, ForeignKey

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase
from vortex.Tuple import Tuple, addTupleType

logger = logging.getLogger(__name__)


@addTupleType
class InternalUserGroupTuple(Tuple, DeclarativeBase):
    """ Internal User Group Tuple

    This table stores the relationships between the users and the groups.

    """
    __tupleType__ = userPluginTuplePrefix + 'InternalUserGroupTuple'
    __tablename__ = 'InternalUserGroup'

    userId = Column(Integer, ForeignKey('InternalUser.id', ondelete='CASCADE')
                    , primary_key=True, nullable=False)

    groupId = Column(Integer, ForeignKey('InternalGroup.id', ondelete='CASCADE')
                     , primary_key=True, nullable=False)

    __table_args__ = (
        Index("idx_InternalUserGroup_map", userId, groupId, unique=True),
    )



import logging

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_core_user._private.server.api.UserHookApi import UserHookApi
from peek_core_user._private.server.api.UserImportApi import UserImportApi
from peek_core_user._private.server.api.UserInfoApi import UserInfoApi
from peek_core_user._private.server.api.UserLoginApi import UserLoginApi
from peek_core_user._private.server.controller.ImportController import \
    ImportController
from peek_core_user._private.server.controller.LoginLogoutController import \
    LoginLogoutController
from peek_core_user.server.UserApiABC import UserApiABC
from peek_core_user.server.UserHookApiABC import UserHookApiABC
from peek_core_user.server.UserImportApiABC import UserImportApiABC
from peek_core_user.server.UserInfoApiABC import UserInfoApiABC
from peek_core_user.server.UserLoginApiABC import UserLoginApiABC

logger = logging.getLogger(__name__)


class UserApi(UserApiABC):

    def __init__(self, deviceApi: DeviceApiABC,
                 dbSessionCreator,
                 importController: ImportController,
                 loginLogoutController: LoginLogoutController):
        self._hookApi = UserHookApi()

        self._importApi = UserImportApi(importController=importController)

        self._infoApi = UserInfoApi(deviceApi=deviceApi,
                                    dbSessionCreator=dbSessionCreator)

        self._loginApi = UserLoginApi(loginLogoutController=loginLogoutController)

    def shutdown(self):
        self._loginApi.shutdown()
        self._importApi.shutdown()
        self._hookApi.shutdown()
        self._infoApi.shutdown()

        self._loginApi = None
        self._importApi = None
        self._hookApi = None
        self._infoApi = None

    @property
    def loginApi(self) -> UserLoginApi:
        return self._loginApi

    @property
    def importApi(self) -> UserImportApi:
        return self._importApi

    @property
    def hookApi(self) -> UserHookApi:
        return self._hookApi

    @property
    def infoApi(self) -> UserInfoApi:
        return self._infoApi



from typing import Callable

from abc import ABCMeta, abstractmethod
from twisted.internet.defer import Deferred

from peek_core_user.tuples.login.UserLoginAction import UserLoginAction
from peek_core_user.tuples.login.UserLoginResponseTuple import UserLoginResponseTuple
from peek_core_user.tuples.login.UserLogoutAction import UserLogoutAction
from peek_core_user.tuples.login.UserLogoutResponseTuple import UserLogoutResponseTuple

UserPostLoginHookCallable = Callable[[UserLoginResponseTuple], Deferred]

UserPostLogoutHookCallable = Callable[[UserLogoutResponseTuple], Deferred]


class UserLoginApiABC(metaclass=ABCMeta):
    @abstractmethod
    def logout(self,
               logoutTuple: UserLogoutAction) -> Deferred:  # [UserLogoutResponseTuple]:
        """ Logout

        :param logoutTuple
        """

    @abstractmethod
    def login(self, loginTuple: UserLoginAction) -> Deferred:  # [UserLoginResponseTuple]:
        """ Login

        :param loginTuple
        """



"""initial

Peek Plugin Database Migration Script

Revision ID: 3fec57d8da58
Revises: 
Create Date: 2017-07-13 22:44:36.374877

"""

# revision identifiers, used by Alembic.
revision = '3fec57d8da58'
down_revision = None
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('UserLoggedIn',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('loggedInDateTime', sa.DateTime(), nullable=False),
    sa.Column('userId', sa.String(length=50), nullable=True),
    sa.Column('clientDeviceToken', sa.String(length=100), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('clientDeviceToken'),
    sa.UniqueConstraint('userId'),
    schema='core_user'
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('UserLoggedIn', schema='core_user')
    # ### end Alembic commands ###



from typing import Optional, List, Dict

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from vortex.Tuple import addTupleType, Tuple, TupleField


@addTupleType
class UserDetailTuple(Tuple):
    __tupleType__ = userPluginTuplePrefix + "UserDetailTuple"

    #:  The username / userid of the user, EG C917
    userName: str = TupleField()

    #:  The title of the user, EG 'Chief Wiggum'
    userTitle: str = TupleField()

    #:  An external system user uuid, EG 715903a7ebc14fb0afb00d432676c51c
    userUuid: Optional[str] = TupleField()

    #:  The mobile number, EG +61 419 123 456
    mobile: Optional[str] = TupleField()

    #:  The email address, EG guy@place.com
    email: Optional[str] = TupleField()

    #:  A list of group names that this user belongs to
    groupNames: List[str] = TupleField()

    #: A field for additional data
    data: Optional[Dict] = TupleField()



import logging
import os

from celery import Celery
from jsoncfg.value_mappers import require_string
from sqlalchemy import MetaData

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_plugin_base.server.PluginServerEntryHookABC import PluginServerEntryHookABC
from peek_plugin_base.server.PluginServerStorageEntryHookABC import \
    PluginServerStorageEntryHookABC
from peek_plugin_base.server.PluginServerWorkerEntryHookABC import \
    PluginServerWorkerEntryHookABC
from peek_core_user._private.server.AdminTupleDataObservable import \
    makeAdminTupleDataObservableHandler
from peek_core_user._private.server.ClientTupleActionProcessor import \
    makeTupleActionProcessorHandler
from peek_core_user._private.server.ClientTupleDataObservable import \
    makeTupleDataObservableHandler
from peek_core_user._private.server.admin_backend import makeAdminBackendHandlers
from peek_core_user._private.server.api.UserApi import UserApi
from peek_core_user._private.server.controller.ImportController import \
    ImportController
from peek_core_user._private.server.controller.LoginLogoutController import \
    LoginLogoutController
from peek_core_user._private.server.controller.MainController import MainController
from peek_core_user._private.tuples.LoggedInUserStatusTuple import \
    LoggedInUserStatusTuple
from peek_core_user.server.UserApiABC import UserApiABC
from peek_plugin_base.storage.DbConnection import DbConnection

logger = logging.getLogger(__name__)


class PluginServerEntryHook(PluginServerEntryHookABC,
                            PluginServerStorageEntryHookABC,
                            PluginServerWorkerEntryHookABC):
    def __init__(self, *args, **kwargs):
        PluginServerEntryHookABC.__init__(self, *args, **kwargs)
        self._userApi = None

        self._handlers = []

    def _migrateStorageSchema(self, metadata: MetaData) -> None:
        """ Migrate Storage Schema

        Rename the schema

        """

        relDir = self._packageCfg.config.storage.alembicDir(require_string)
        alembicDir = os.path.join(self.rootDir, relDir)
        if not os.path.isdir(alembicDir): raise NotADirectoryError(alembicDir)

        dbConn = DbConnection(
            dbConnectString=self.platform.dbConnectString,
            metadata=metadata,
            alembicDir=alembicDir,
            enableCreateAll=False
        )

        # Rename the plugin schema to core.
        renameToCoreSql = '''
            DO $$
            BEGIN
                IF EXISTS(
                    SELECT schema_name
                      FROM information_schema.schemata
                      WHERE schema_name = 'pl_user'
                  )
                THEN
                  EXECUTE ' DROP SCHEMA IF EXISTS core_user CASCADE ';
                  EXECUTE ' ALTER SCHEMA pl_user RENAME TO core_user ';
                END IF;
            END
            $$;
        '''

        dbSession = dbConn.ormSessionCreator()
        try:
            dbSession.execute(renameToCoreSql)
            dbSession.commit()

        finally:
            dbSession.close()
            dbConn.dbEngine.dispose()

        PluginServerStorageEntryHookABC._migrateStorageSchema(self, metadata)

    def load(self) -> None:
        from peek_core_user._private.tuples import loadPrivateTuples
        loadPrivateTuples()

        from peek_core_user._private.storage.DeclarativeBase import loadStorageTuples
        loadStorageTuples()

        from peek_core_user.tuples import loadPublicTuples
        loadPublicTuples()

        logger.debug("loaded")

    def start(self):
        """ Start

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """
        # ----------------
        # Setup the APIs
        deviceApi: DeviceApiABC = self.platform.getOtherPluginApi("peek_core_device")

        # ----------------
        # Import Controller
        importController = ImportController()
        self._handlers.append(importController)

        # ----------------
        # Login / Logout Controller
        loginLogoutController = LoginLogoutController(deviceApi, self.dbSessionCreator)
        self._handlers.append(importController)

        # ----------------
        # Setup our API
        self._userApi = UserApi(deviceApi,
                                self.dbSessionCreator,
                                importController,
                                loginLogoutController)

        # ----------------
        # Main Controller
        mainController = MainController(self.dbSessionCreator, self._userApi)
        self._handlers.append(mainController)

        # ----------------
        # Client Tuple Observable
        clientTupleObservable = makeTupleDataObservableHandler(
            self.dbSessionCreator, self._userApi
        )
        self._handlers.append(clientTupleObservable)

        # ----------------
        # Admin Tuple Observable
        adminTupleObservable = makeAdminTupleDataObservableHandler(
            self.dbSessionCreator, deviceApi, self._userApi
        )
        self._handlers.append(clientTupleObservable)

        # ----------------
        # Setup controllers.
        loginLogoutController.setup(clientTupleObservable,
                                    adminTupleObservable,
                                    self._userApi.hookApi,
                                    self._userApi.infoApi)
        importController.setTupleObserver(clientTupleObservable)

        # Make the admin observable send an update when device online / offline
        # state changes occur
        deviceApi.deviceOnlineStatus().subscribe(
            lambda _: adminTupleObservable.notifyOfTupleUpdateForTuple(
                LoggedInUserStatusTuple.tupleType()
            )
        )

        # ----------------
        # Setup the Action Processor
        self._handlers.append(makeTupleActionProcessorHandler(mainController))

        # ----------------
        # Setup admin backend
        self._handlers.extend(
            makeAdminBackendHandlers(clientTupleObservable, self.dbSessionCreator)
        )

        logger.debug("started")

    def stop(self):
        for handler in self._handlers:
            handler.shutdown()

        self._userApi.shutdown()

        logger.debug("stopped")

    def unload(self):
        logger.debug("unloaded")

    @property
    def publishedServerApi(self) -> UserApiABC:
        return self._userApi

    @property
    def dbMetadata(self):
        from peek_core_user._private.storage import DeclarativeBase
        return DeclarativeBase.metadata

    ###### Implement PluginServerWorkerEntryHookABC

    @property
    def celeryApp(self) -> Celery:
        from peek_core_user._private.worker.CeleryApp import celeryApp
        return celeryApp



import logging

from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler
from vortex.sqla_orm.OrmCrudHandler import OrmCrudHandler, OrmCrudHandlerExtension

from peek_core_user._private.PluginNames import userPluginFilt
from peek_core_user._private.storage.Setting import SettingProperty, globalSetting, \
    ldapSetting
from peek_core_user._private.tuples.UserLoginUiSettingTuple import UserLoginUiSettingTuple

logger = logging.getLogger(__name__)

# This dict matches the definition in the Admin angular app.
filtKey = {"key": "admin.Edit.SettingProperty"}
filtKey.update(userPluginFilt)


# This is the CRUD handler
class __CrudHandler(OrmCrudHandler):
    # The UI only edits the global settings
    # You could get more complicated and have the UI edit different groups of settings.
    def createDeclarative(self, session, payloadFilt):
        settingType = payloadFilt.get('settingType')
        if settingType == "Global":
            return [p for p in globalSetting(session).propertyObjects]
        elif settingType == 'LDAP':
            return [p for p in ldapSetting(session).propertyObjects]

        raise Exception("%s is not a known settings group" % settingType)


class __ExtUpdateObservable(OrmCrudHandlerExtension):
    """ Update Observable ORM Crud Extension

    This extension is called after events that will alter data,
    it then notifies the observer.

    """

    def __init__(self, tupleDataObserver: TupleDataObservableHandler):
        self._tupleDataObserver = tupleDataObserver

    def _afterCommit(self, tuple_, tuples, session, payloadFilt):
        self._tupleDataObserver.notifyOfTupleUpdate(
            TupleSelector(UserLoginUiSettingTuple.tupleName(), {})
        )
        return True

    afterUpdateCommit = _afterCommit
    afterDeleteCommit = _afterCommit


# This method creates an instance of the handler class.
def makeSettingPropertyHandler(tupleObservable, dbSessionCreator):
    handler = __CrudHandler(dbSessionCreator, SettingProperty,
                            filtKey, retreiveAll=True)
    handler.addExtension(SettingProperty, __ExtUpdateObservable(tupleObservable))

    logger.debug("Started")
    return handler



import logging
from typing import List

from twisted.cred.error import LoginFailed

from peek_core_user._private.server.controller.PasswordUpdateController import \
    PasswordUpdateController
from peek_core_user._private.storage.InternalGroupTuple import InternalGroupTuple
from peek_core_user._private.storage.InternalUserGroupTuple import InternalUserGroupTuple
from peek_core_user._private.storage.InternalUserPassword import InternalUserPassword
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user.server.UserDbErrors import UserPasswordNotSetException

logger = logging.getLogger(__name__)


class InternalAuth:

    def checkPassBlocking(self, dbSession, userName, password) -> List[str]:

        results = dbSession.query(InternalUserPassword) \
            .join(InternalUserTuple) \
            .filter(InternalUserTuple.userName == userName) \
            .all()

        if not results or not results[0].password:
            raise UserPasswordNotSetException(userName)

        passObj = results[0]
        if passObj.password != PasswordUpdateController.hashPass(password):
            raise LoginFailed("Username or password is incorrect")

        groups = dbSession.query(InternalGroupTuple) \
            .join(InternalUserGroupTuple) \
            .filter(InternalUserGroupTuple.userId == passObj.userId) \
            .all()

        return [g.groupName for g in groups]




from vortex.handler.TupleActionProcessorProxy import TupleActionProcessorProxy

from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_core_user._private.PluginNames import userPluginFilt, \
    userPluginActionProcessorName


def makeTupleActionProcessorProxy():
    return TupleActionProcessorProxy(
        tupleActionProcessorName=userPluginActionProcessorName,
        proxyToVortexName=peekServerName,
        additionalFilt=userPluginFilt)



import logging
from datetime import datetime
from typing import List

import pytz
from twisted.cred.error import LoginDenied, LoginFailed
from twisted.internet.defer import Deferred, inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_core_user._private.server.api.UserHookApi import UserHookApi
from peek_core_user._private.server.api.UserInfoApi import UserInfoApi
from peek_core_user._private.server.auth_connectors.InternalAuth import InternalAuth
from peek_core_user._private.server.auth_connectors.LdapAuth import LdapAuth
from peek_core_user._private.storage.Setting import ldapSetting, LDAP_ENABLED
from peek_core_user._private.storage.UserLoggedIn import UserLoggedIn
from peek_core_user._private.tuples.LoggedInUserStatusTuple import \
    LoggedInUserStatusTuple
from peek_core_user._private.tuples.UserLoggedInTuple import UserLoggedInTuple
from peek_core_user.server.UserDbErrors import UserIsNotLoggedInToThisDeviceError
from peek_core_user.tuples.login.UserLoginAction import UserLoginAction
from peek_core_user.tuples.login.UserLoginResponseTuple import UserLoginResponseTuple
from peek_core_user.tuples.login.UserLogoutAction import UserLogoutAction
from peek_core_user.tuples.login.UserLogoutResponseTuple import UserLogoutResponseTuple
from peek_plugin_base.storage.DbConnection import DbSessionCreator

logger = logging.getLogger(__name__)

USER_ALREADY_LOGGED_ON_KEY = 'pl-user.USER_ALREADY_LOGGED_ON'
DEVICE_ALREADY_LOGGED_ON_KEY = 'pl-user.DEVICE_ALREADY_LOGGED_ON_KEY'


class LoginLogoutController:

    def __init__(self, deviceApi: DeviceApiABC,
                 dbSessionCreator: DbSessionCreator):
        self._deviceApi: DeviceApiABC = deviceApi
        self._hookApi: UserHookApi = None
        self._infoApi: UserInfoApi = None
        self._dbSessionCreator: DbSessionCreator = dbSessionCreator
        self._clientTupleObservable: TupleDataObservableHandler = None
        self._adminTupleObservable: TupleDataObservableHandler = None

    def setup(self, clientTupleObservable,
              adminTupleObservable,
              hookApi: UserHookApi,
              infoApi: UserInfoApi):
        self._clientTupleObservable = clientTupleObservable
        self._adminTupleObservable = adminTupleObservable
        self._hookApi = hookApi
        self._infoApi = infoApi

    def shutdown(self):
        self._clientTupleObservable = None
        self._hookApi = None
        self._infoApi = None

    def _checkPassBlocking(self, ormSession, userName, password) -> List[str]:
        if not password:
            raise LoginFailed("Password is empty")

        if ldapSetting(ormSession, LDAP_ENABLED):
            return LdapAuth().checkPassBlocking(ormSession, userName, password)

        return InternalAuth().checkPassBlocking(ormSession, userName, password)

    def _checkGroupBlocking(self, ormSession, groups:List[str]):
        pass

    @deferToThreadWrapWithLogger(logger)
    def _logoutInDb(self, logoutTuple: UserLogoutAction):
        """
        Returns Deferred[UserLogoutResponseTuple]
        """

        session = self._dbSessionCreator()
        try:
            # Check if the user is actually logged into this device.
            qry = (
                session.query(UserLoggedIn)
                    .filter(UserLoggedIn.userName == logoutTuple.userName)
                    .filter(
                    UserLoggedIn.deviceToken == logoutTuple.deviceToken)
            )

            if qry.count() == 0:
                raise UserIsNotLoggedInToThisDeviceError(logoutTuple.userName)

            session.delete(qry.one())
            session.commit()

        finally:
            session.close()

    @inlineCallbacks
    def logout(self, logoutTuple: UserLogoutAction) -> Deferred:
        """ Logout

        :param logoutTuple: The tuple containing the information to process
                                for the logout.

        :return A deferred that fires with List[UserLogoutResponseTuple]
        """

        deviceDescription = yield self._deviceApi.deviceDescription(
            logoutTuple.deviceToken
        )

        response = UserLogoutResponseTuple(
            userName=logoutTuple.userName,
            deviceToken=logoutTuple.deviceToken,
            deviceDescription=deviceDescription,
            acceptedWarningKeys=logoutTuple.acceptedWarningKeys,
            succeeded=True)

        # Give the hooks a chance to fail the logout
        yield self._hookApi.callLogoutHooks(response)

        # If there are no problems, proceed with the logout.
        if response.succeeded:
            yield self._logoutInDb(logoutTuple)

        self._clientTupleObservable.notifyOfTupleUpdate(
            TupleSelector(UserLoggedInTuple.tupleType(),
                          selector=dict(userName=logoutTuple.userName))
        )

        self._adminTupleObservable.notifyOfTupleUpdateForTuple(
            LoggedInUserStatusTuple.tupleType()
        )

        return response

    @deferToThreadWrapWithLogger(logger)
    def _loginInDb(self, loginTuple: UserLoginAction, singleDevice: bool = True):
        """
        Returns Deferred[UserLoginResponseTuple]

        """

        assert singleDevice, "Multiple device login not implemented"
        userName = loginTuple.userName
        password = loginTuple.password
        acceptedWarningKeys = set(loginTuple.acceptedWarningKeys)
        deviceToken = loginTuple.deviceToken
        vehicle = loginTuple.vehicleId

        responseTuple = UserLoginResponseTuple(
            userName=userName,
            userToken="Not implemented",
            succeeded=False,
            acceptedWarningKeys=loginTuple.acceptedWarningKeys,
            vehicleId=loginTuple.vehicleId
        )

        if not deviceToken:
            raise Exception("peekToken must be supplied")

        thisDeviceDescription = self._deviceApi.deviceDescriptionBlocking(deviceToken)

        ormSession = self._dbSessionCreator()
        try:
            groups = self._checkPassBlocking(ormSession, userName, password)
            self._checkGroupBlocking(ormSession, groups)

            responseTuple.userDetail = self._infoApi.userBlocking(userName, ormSession)

            # Find any current login sessions
            userLoggedIn = (ormSession.query(UserLoggedIn)
                            .filter(UserLoggedIn.userName == userName)
                            .all())
            userLoggedIn = userLoggedIn[0] if userLoggedIn else None

            loggedInElsewhere = (
                ormSession.query(UserLoggedIn)
                    .filter(UserLoggedIn.deviceToken != deviceToken)
                    .filter(UserLoggedIn.userName == userName)
                    .all())

            if len(loggedInElsewhere) not in (0, 1):
                raise Exception("Found more than 1 ClientDevice for"
                                + (" token %s" % deviceToken))

            loggedInElsewhere = loggedInElsewhere[0] if loggedInElsewhere else None

            sameDevice = userLoggedIn and loggedInElsewhere is None

            # If the user is logged in, but not to this client device, raise exception
            if userLoggedIn and not sameDevice:
                if USER_ALREADY_LOGGED_ON_KEY in acceptedWarningKeys:
                    self._forceLogout(ormSession,
                                      userName,
                                      loggedInElsewhere.deviceToken)
                    userLoggedIn = False

                else:
                    otherDeviceDescription = self._deviceApi.deviceDescriptionBlocking(
                        loggedInElsewhere.deviceToken
                    )

                    # This is false if the logged in device has been removed from
                    # enrollment
                    if otherDeviceDescription:
                        responseTuple.setFailed()
                        responseTuple.addWarning(
                            USER_ALREADY_LOGGED_ON_KEY,
                            "User %s is already logged in, on device %s" % (
                                userName, otherDeviceDescription)
                        )

                        return responseTuple

                    # Else, The old device has been deleted,
                    # Just let them login to the same device.
                    self._forceLogout(ormSession,
                                      loggedInElsewhere.userName,
                                      loggedInElsewhere.deviceToken)

            # If we're logging into the same device, but already logged in
            if sameDevice:  # Logging into the same device
                sameDeviceDescription = self._deviceApi.deviceDescriptionBlocking(
                    userLoggedIn.deviceToken
                )

                responseTuple.deviceToken = userLoggedIn.deviceToken
                responseTuple.deviceDescription = sameDeviceDescription
                responseTuple.succeeded = True
                return responseTuple

            anotherUserOnThatDevice = (
                ormSession.query(UserLoggedIn)
                    .filter(UserLoggedIn.deviceToken == deviceToken)
                    .filter(UserLoggedIn.userName != userName)
                    .all()
            )

            if anotherUserOnThatDevice:
                anotherUserOnThatDevice = anotherUserOnThatDevice[0]
                if DEVICE_ALREADY_LOGGED_ON_KEY in acceptedWarningKeys:
                    self._forceLogout(ormSession,
                                      anotherUserOnThatDevice.userName,
                                      anotherUserOnThatDevice.deviceToken)

                else:
                    responseTuple.setFailed()
                    responseTuple.addWarning(
                        DEVICE_ALREADY_LOGGED_ON_KEY,
                        "User %s is currently logged into this device : %s" % (
                            anotherUserOnThatDevice.userName,
                            thisDeviceDescription)
                    )

                    return responseTuple

            # Create the user logged in entry

            newUser = UserLoggedIn(userName=userName,
                                   loggedInDateTime=datetime.now(pytz.utc),
                                   deviceToken=deviceToken,
                                   vehicle=vehicle)
            ormSession.add(newUser)
            ormSession.commit()

            # Respond with a successful login
            responseTuple.deviceToken = deviceToken
            responseTuple.deviceDescription = thisDeviceDescription
            responseTuple.succeeded = True
            return responseTuple

        finally:
            ormSession.close()

    @inlineCallbacks
    def login(self, loginTuple: UserLoginAction):
        """
        Returns Deferred[UserLoginResponseTuple]

        """
        loginResponse = None
        try:
            loginResponse = yield self._loginInDb(loginTuple)
            yield self._hookApi.callLoginHooks(loginResponse)

        # except UserAlreadyLoggedInError as e:
        #     pass
        #
        # except DeviceAlreadyLoggedInError as e:
        #     pass
        #
        # except UserIsNotLoggedInToThisDeviceError as e:
        #     pass

        except Exception as e:

            # Log the user out again if the hooks fail
            logoutTuple = UserLogoutAction(
                userName=loginTuple.userName,
                deviceToken=loginTuple.deviceToken)

            # Force logout, we don't care if it works or not.
            try:
                yield self._logoutInDb(logoutTuple)
            except UserIsNotLoggedInToThisDeviceError:
                pass

            raise e

        self._clientTupleObservable.notifyOfTupleUpdate(
            TupleSelector(UserLoggedInTuple.tupleType(),
                          selector=dict(userName=loginTuple.userName))
        )

        self._adminTupleObservable.notifyOfTupleUpdateForTuple(
            LoggedInUserStatusTuple.tupleType()
        )

        return loginResponse

    def _forceLogout(self, ormSession, userName, deviceToken):
        (
            ormSession.query(UserLoggedIn)
                .filter(UserLoggedIn.userName == userName)
                .filter(UserLoggedIn.deviceToken == deviceToken)
                .delete(synchronize_session=False)
        )



import logging

from twisted.internet.defer import Deferred, inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.storage.UserLoggedIn import UserLoggedIn
from peek_core_user._private.tuples.LoggedInUserStatusTuple import \
    LoggedInUserStatusTuple

logger = logging.getLogger(__name__)


class LoggedInUserStatusTupleProvider(TuplesProviderABC):
    def __init__(self, dbSessionCreator,
                 deviceApi: DeviceApiABC):
        self._dbSessionCreator = dbSessionCreator
        self._deviceApi = deviceApi

        assert isinstance(self._deviceApi, DeviceApiABC), (
            "We didn't get a DeviceApiABC")

    @inlineCallbacks
    def makeVortexMsg(self, filt: dict, tupleSelector: TupleSelector) -> Deferred:

        tuples = yield self._loadLoggedInTuples()
        tuplesByDeviceToken = {t.deviceToken: t for t in tuples}

        deviceTokens = list(tuplesByDeviceToken)

        deviceDetails = yield self._deviceApi.deviceDetails(deviceTokens)

        for deviceDetail in deviceDetails:
            tuple_ = tuplesByDeviceToken[deviceDetail.deviceToken]
            tuple_.deviceIsOnline = deviceDetail.isOnline
            tuple_.deviceLastOnline = deviceDetail.lastOnline
            tuple_.deviceType = deviceDetail.deviceType
            tuple_.deviceDescription = deviceDetail.description

        payloadEnvelope = yield Payload(filt=filt, tuples=tuples) \
            .makePayloadEnvelopeDefer()
        vortexMsg = yield payloadEnvelope.toVortexMsgDefer()
        return vortexMsg

    @deferToThreadWrapWithLogger(logger)
    def _loadLoggedInTuples(self) -> Deferred:
        dbSession = self._dbSessionCreator()
        try:
            tuples = [
                LoggedInUserStatusTuple(
                    userName=u.userName,
                    userTitle=u.userTitle,
                    loginDate=u.loggedInDateTime,
                    vehicle=u.vehicle,
                    deviceToken=u.deviceToken
                )
                for u in dbSession.query(UserLoggedIn.userName,
                                         UserLoggedIn.loggedInDateTime,
                                         UserLoggedIn.deviceToken,
                                         UserLoggedIn.vehicle,
                                         InternalUserTuple.userTitle)
                    .join(InternalUserTuple,
                          UserLoggedIn.userName == InternalUserTuple.userName)
            ]

            return tuples

        finally:
            dbSession.close()



"""separated password

Peek Plugin Database Migration Script

Revision ID: 2c6728ebc562
Revises: 9b65be31926e
Create Date: 2018-01-28 16:39:17.645627

"""

# revision identifiers, used by Alembic.
revision = '2c6728ebc562'
down_revision = '9b65be31926e'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():

    op.create_table('InternalUserPassword',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('userId', sa.Integer(), nullable=False),
    sa.Column('password', sa.String(length=50), nullable=False),
    sa.ForeignKeyConstraint(['userId'], ['core_user.InternalUser.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    schema='core_user'
    )

    op.execute('''
    INSERT INTO core_user."InternalUserPassword" ("userId", "password")
    SELECT "id", "password"
    FROM core_user."InternalUser"
    WHERE "password" is not null
    ''')

    op.drop_column('InternalUser', 'password', schema='core_user')



def downgrade():
    raise NotImplementedError("Downgrade not implemented")



import logging

from twisted.internet.defer import Deferred

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_core_user._private.server.controller.LoginLogoutController import \
    LoginLogoutController
from peek_core_user.server.UserLoginApiABC import UserLoginApiABC
from peek_core_user.tuples.login.UserLoginAction import UserLoginAction
from peek_core_user.tuples.login.UserLogoutAction import UserLogoutAction

logger = logging.getLogger(__name__)


class UserLoginApi(UserLoginApiABC):
    #: A reference to the core device plugins API
    _deviceApi: DeviceApiABC

    def __init__(self, loginLogoutController:LoginLogoutController):
        self._loginLogoutController = loginLogoutController

    def shutdown(self):
        pass

    def logout(self, logoutTuple: UserLogoutAction) -> Deferred:
        return self._loginLogoutController.logout(logoutTuple)


    def login(self, loginTuple: UserLoginAction) -> Deferred:
        return self._loginLogoutController.login(loginTuple)


import logging
from typing import List

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from vortex.Tuple import addTupleType, Tuple, TupleField

logger = logging.getLogger(__name__)


@addTupleType
class UserLoggedInTuple(Tuple):
    """ User Logged In Tuple

      This tuple is sent to the devices when a user logs in.

      If the device receives this tuple and the deviceToken doesn't match the current
      device, then the user is logged off.

    """
    __tupleType__ = userPluginTuplePrefix + "UserLoggedInTuple"

    userName: str = TupleField()
    deviceToken: str = TupleField()




import logging

from twisted.internet.defer import Deferred, inlineCallbacks

from peek_core_user._private.server.api.UserApi import UserApi
from peek_core_user._private.tuples.UserLoggedInTuple import UserLoggedInTuple
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)


class UserLoggedInTupleProvider(TuplesProviderABC):
    def __init__(self, dbSessionCreator,
                 ourApi: UserApi):
        self._dbSessionCreator = dbSessionCreator
        self._ourApi = ourApi

        from peek_core_user.server.UserApiABC import UserApiABC
        assert isinstance(self._ourApi, UserApiABC), (
            "We didn't get a UserApiABC")

    @inlineCallbacks
    def makeVortexMsg(self, filt: dict, tupleSelector: TupleSelector) -> Deferred:
        userName = tupleSelector.selector["userName"]

        deviceToken = yield self._ourApi.infoApi.peekDeviceTokenForUser(userName)

        tuple_ = UserLoggedInTuple(userName=userName, deviceToken=deviceToken)

        payload = Payload(filt=filt, tuples=[tuple_])
        payloadEnvelope = yield payload.makePayloadEnvelopeDefer()
        vortexMsg = yield payloadEnvelope.toVortexMsgDefer()
        return vortexMsg





# class UserIsNotLoggedInError(Exception):
#     def __init__(self, userName):
#         self.userName = userName
#
#     def __str__(self):
#         return "User %s is not logged in on any device" % self.userName


class UserIsNotLoggedInToThisDeviceError(Exception):
    def __init__(self, userName):
        self.userName = userName

    def __str__(self):
        return "User %s is not logged into this device" % self.userName


class UserNotFoundException(Exception):
    def __init__(self, userName):
        self.userName = userName

    def __str__(self):
        if self.userName is None:
            return "User name not selected"
        else:
            return "User %s is not found in database" % self.userName



class UserPasswordNotSetException(Exception):
    def __init__(self, userName):
        self.userName = userName

    def __str__(self):
        if self.userName is None:
            return "UserPasswordNotSetException:User name not selected"
        else:
            return "No password set for user %s" % self.userName



from typing import Callable

from abc import ABCMeta, abstractmethod
from rx.subjects import Subject
from twisted.internet.defer import Deferred

from peek_core_user.tuples.login.UserLoginResponseTuple import UserLoginResponseTuple
from peek_core_user.tuples.login.UserLogoutResponseTuple import UserLogoutResponseTuple

UserPostLoginHookCallable = Callable[[UserLoginResponseTuple], Deferred]

UserPostLogoutHookCallable = Callable[[UserLogoutResponseTuple], Deferred]


class UserHookApiABC(metaclass=ABCMeta):
    @abstractmethod
    def postLoginObservable(self) -> Subject:
        """ Get the post login observable

        Subscribers will be called just after the user has authenticated

        """

    @abstractmethod
    def postLogoffObservable(self) -> Subject:
        """ Get the post logoff observable

        Subscribers will be called just after the user has logged off

        """

    @abstractmethod
    def addPostLoginHook(self, callable: UserPostLoginHookCallable) -> None:
        """ Add Post Login Hook

        :param callable: This callable will be called just after the user has
        authenticated
        """

    @abstractmethod
    def removePostLoginHook(self, callable: UserPostLoginHookCallable) -> None:
        """ Remove Post Login Hook

        :param callable: This callable to remove from the hooks
        """

    @abstractmethod
    def addPostLogoutHook(self, callable: UserPostLogoutHookCallable) -> None:
        """ Add Post Logout Hook

        :param callable: This callable will be called just after the user has
        logged out
        """

    @abstractmethod
    def removePostLogoutHook(self, callable: UserPostLogoutHookCallable) -> None:
        """ Remove Post Logout Hook

        :param callable: This callable to remove from the hooks
        """



import logging
from typing import List

from twisted.cred.error import LoginFailed

from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple

__author__ = 'synerty'

logger = logging.getLogger(__name__)

import ldap

'''
{'objectClass': [b'top', b'person', b'organizationalPerson', b'user'], 'cn': [b'attest'],
 'givenName': [b'attest'],
 'distinguishedName': [b'CN=attest,OU=testou,DC=synad,DC=synerty,DC=com'],
 'instanceType': [b'4'], 'whenCreated': [b'20170505160836.0Z'],
 'whenChanged': [b'20190606130621.0Z'], 'displayName': [b'attest'],
 'uSNCreated': [b'16498'],
 'memberOf': [b'CN=Domain Admins,CN=Users,DC=synad,DC=synerty,DC=com',
              b'CN=Enterprise Admins,CN=Users,DC=synad,DC=synerty,DC=com',
              b'CN=Administrators,CN=Builtin,DC=synad,DC=synerty,DC=com'],
 'uSNChanged': [b'73784'], 'name': [b'attest'],
 'objectGUID': [b'\xee\x1bV\x8dQ\xackE\x82\xd9%_\x18\xadjO'],
 'userAccountControl': [b'66048'], 'badPwdCount': [b'0'], 'codePage': [b'0'],
 'countryCode': [b'0'], 'badPasswordTime': [b'132042996316396717'], 'lastLogoff': [b'0'],
 'lastLogon': [b'132042996806397639'], 'pwdLastSet': [b'132042997225927009'],
 'primaryGroupID': [b'513'], 'objectSid': [
    b'\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00D:3|X\x8f\xc7\x08\xe6\xeaV\xc8Q\x04\x00\x00'],
 'adminCount': [b'1'], 'accountExpires': [b'9223372036854775807'], 'logonCount': [b'36'],
 'sAMAccountName': [b'attest'], 'sAMAccountType': [b'805306368'],
 'userPrincipalName': [b'attest@synad.synerty.com'], 'lockoutTime': [b'0'],
 'objectCategory': [b'CN=Person,CN=Schema,CN=Configuration,DC=synad,DC=synerty,DC=com'],
 'dSCorePropagationData': [b'20190606130621.0Z', b'20190606130016.0Z',
                           b'20170506090346.0Z', b'16010101000000.0Z'],
 'lastLogonTimestamp': [b'132042996806397639']}
'''

class LdapNotEnabledError(Exception):
    pass


class LdapAuth:

    def checkPassBlocking(self, dbSession, userName, password) -> List[str]:
        """ Login User

        :param dbSession:
        :param userName: The username of the user.
        :param password: The users secret password.
        :rtype
        """
        (ldapDoamin, ldapOuFolders, ldapCnFolders, ldapUri) = self._loadLdapSettings(
            dbSession
        )

        try:

            conn = ldap.initialize(ldapUri)
            conn.protocol_version = 3
            conn.set_option(ldap.OPT_REFERRALS, 0)

            # make the connection
            conn.simple_bind_s('%s@%s' % (userName, ldapDoamin), password)
            ldapFilter = "(&(objectCategory=person)(objectClass=user)(sAMAccountName=%s))" % userName

            dcParts = ','.join(['DC=%s' % part for part in ldapDoamin.split('.')])

            ldapBases = self._makeLdapBase(ldapOuFolders, userName, "OU")
            ldapBases += self._makeLdapBase(ldapCnFolders, userName, "CN")

            for ldapBase in ldapBases:
                ldapBase = "%s,%s" % (ldapBase, dcParts)

                try:
                    # Example Base : 'CN=atuser1,CN=Users,DC=synad,DC=synerty,DC=com'
                    userDetails = conn.search_st(ldapBase, ldap.SCOPE_SUBTREE,
                                                 ldapFilter, None, 0, 10)

                    if userDetails:
                        break

                except ldap.NO_SUCH_OBJECT:
                    raise

        except ldap.NO_SUCH_OBJECT:
            logger.error("Login failed for %s, failed to query LDAP for user groups",
                         userName)
            raise LoginFailed(
                "An internal error occurred, ask admin to check Attune logs")

        except ldap.INVALID_CREDENTIALS:
            logger.error("Login failed for %s, invalid credentials", userName)
            raise LoginFailed("Username or password is incorrect")

        if not userDetails:
            logger.error("Login failed for %s, failed to query LDAP for user groups",
                         userName)
            raise LoginFailed(
                "An internal error occurred, ask admin to check Attune logs")

        userDetails = userDetails[0][1]

        groups = []
        for memberOf in userDetails['memberOf']:
            group = memberOf.decode().split(',')[0]
            if '=' in group:
                group = group.split('=')[1]
            groups.append(group)

        userTitle = None
        if userDetails['displayName']:
            userTitle = userDetails['displayName'][0].decode()

        email = None
        if userDetails['userPrincipalName']:
            email = userDetails['userPrincipalName'][0].decode()

        userUuid = None
        if userDetails['distinguishedName']:
            userUuid = userDetails['distinguishedName'][0].decode()

        self._makeOrCreateInternalUserBlocking(dbSession,
                                               userName, userTitle, userUuid, email)

        return groups

    def _makeOrCreateInternalUserBlocking(self, dbSession,
                                          userName, userTitle, userUuid, email):

        internalUser = dbSession.query(InternalUserTuple) \
            .filter(InternalUserTuple.userName == userName) \
            .all()

        if internalUser:
            return

        newInternalUser = InternalUserTuple(
            userName=userName,
            userTitle=userTitle,
            userUuid=userUuid,
            email=email
        )

        dbSession.add(newInternalUser)
        dbSession.commit()

    def _loadLdapSettings(self, dbSession):

        from peek_core_user._private.storage.Setting import ldapSetting, \
            LDAP_DOMAIN_NAME, LDAP_URI, LDAP_CN_FOLDERS, \
            LDAP_OU_FOLDERS

        try:
            ldapSettings = ldapSetting(dbSession)

            ldapDoamin = ldapSettings[LDAP_DOMAIN_NAME]
            ldapCnFolders = ldapSettings[LDAP_CN_FOLDERS]
            ldapOuFolders = ldapSettings[LDAP_OU_FOLDERS]
            ldapUri = ldapSettings[LDAP_URI]

        except Exception as e:
            logger.error("Failed to query for LDAP connection settings")
            logger.exception(e)
            raise LoginFailed(
                "An internal error occurred, ask admin to check Attune logs")

        return ldapDoamin, ldapOuFolders, ldapCnFolders, ldapUri

    def _makeLdapBase(self, ldapFolders, userName, propertyName):
        try:
            ldapBases = []
            for folder in ldapFolders.split(','):
                folder = folder.strip()
                if not folder:
                    continue

                parts = []
                for part in folder.split('/'):
                    part = part.strip()
                    if not part:
                        continue
                    parts.append('%s=%s' % (propertyName, part))

                ldapBases.append(','.join(reversed(parts)))

            return ldapBases

        except Exception as e:
            logger.error(
                "Login failed for %s, failed to parse LDAP %s Folders setting" % propertyName,
                userName)

            logger.exception(e)

            raise LoginFailed(
                "An internal error occurred, ask admin to check Attune logs")



import logging

from vortex.Tuple import addTupleType, Tuple, TupleField

from peek_core_user._private.PluginNames import userPluginTuplePrefix

logger = logging.getLogger(__name__)


@addTupleType
class UserLoginUiSettingTuple(Tuple):
    """ User Login UI Setting

      This tuple is sent to the devices to control the appearance of the UI

    """
    __tupleType__ = userPluginTuplePrefix + "UserLoginUiSettingTuple"

    showUsersAsList: bool = TupleField()
    showVehicleInput: bool = TupleField()



import logging
from datetime import datetime

from vortex.Tuple import addTupleType, Tuple, TupleField

from peek_core_user._private.PluginNames import userPluginTuplePrefix

logger = logging.getLogger(__name__)


@addTupleType
class LoggedInUserStatusTuple(Tuple):
    """ Logged In User Status Tuple

      This tuple is used by the "Managed LoggedIn User"

      If the device receives this tuple and the deviceToken doesn't match the current
      device, then the user is logged off.

    """
    __tupleType__ = userPluginTuplePrefix + "LoggedInUserStatusTuple"

    userName: str = TupleField()

    #:  The title of the user, EG 'Chief Wiggum'
    userTitle: str = TupleField()

    #:  The vehicle the user is logged in with
    vehicle: str = TupleField()

    #:  The date that the user logged in
    loginDate: datetime = TupleField()

    #:  The device that the user is logged into From peek-core-device
    deviceToken: str = TupleField()

    #:  Is the device online now
    deviceIsOnline: bool = TupleField()

    #:  The last time the user was online
    deviceLastOnline: datetime = TupleField()

    #:  The device type
    deviceType: str = TupleField()

    #:  The device description
    deviceDescription: str = TupleField()
